// -*- mode:c++ -*-
//
// Header file CodeGen.hh
//
// Copyright (c) 2002 Raoul M. Gough
//
// This material is provided "as is", with absolutely no warranty expressed
// or implied. Any use is at your own risk.
//
// Permission to use or copy this software for any purpose is hereby granted 
// without fee, provided the above notices are retained on all copies.
// Permission to modify the code and to distribute modified code is granted,
// provided the above notices are retained, and a notice that the code was
// modified is included with the above copyright notice.
//
// History
// =======
// 2002/ 2/11	rmg	File creation
//
// version: $Id: CodeGen.hh,v 1.9 2003/09/30 22:42:56 Raoul Exp $
//

#ifndef CodeGen_rmg_20020211_included
#define CodeGen_rmg_20020211_included

#include "clamp_support.hh"

#include <stdio.h>
#include <stack>
#include <memory>
#include <fstream>
#include <map>
#include <sstream>
#include "boost/smart_ptr.hpp"

namespace clamp
{
  //
  // Store lambda expressions for deferred processing, grouping
  // by context and parameter list size (i.e. ccount and pcount)
  //
  struct LambdaCategory
  {
      unsigned ccount;
      unsigned pcount;
      std::string quals;
      
      bool operator==(const LambdaCategory& b) const
      { return ccount==b.ccount
            && pcount==b.pcount
            && quals==b.quals; }
      bool operator< (const LambdaCategory& b) const
      {
        if(ccount!=b.ccount) return ccount<b.ccount;
        if(pcount!=b.pcount) return pcount<b.pcount;
        if(quals !=b.quals)  return quals<b.quals;
        return false;
      }
  };

  class CodeGen
  {
  public:
    CodeGen ();

  public:
    void lambdaType (std::stringstream& , std::auto_ptr<lambdaT>);
    std::auto_ptr<stringT> lambdaTypeName (std::auto_ptr<lambdaT>);

  public:
    bool braceMatch (int pos);
    bool startLambda (std::stringstream*& , int openBracePos, std::auto_ptr<lambdaT>);
    bool endLambda (std::stringstream*& );

  public:
    bool contextRef (std::stringstream& , std::auto_ptr<stringT>, bool byRef = false);

  public:
    void spitItOut (const std::stringstream& program);

  private:
    struct LambdaContext {
      std::stringstream* mPrevFile;
      std::string        mFileName;
      int mOpenBracePos;
      boost::shared_ptr<lambdaT> mLamp;
      int mLambdaNumber;

      static int sContextCount;

      LambdaContext (std::stringstream*, int, std::auto_ptr<lambdaT>);
    };

  private:
    std::stack<LambdaContext> mStack;
    std::string mCurFileName;

  private:
    int mLambdaCount;

  private:
    std::stringstream mImplStream;

  private:
    typedef std::list<LambdaContext> LambdaList;

    struct LambdaGroup
    {
      // Lambda functions that share the same implementation template
      LambdaList mLambdaList;
      boost::shared_ptr<std::stringstream> mImplStream;

      LambdaGroup ();
    };

    typedef std::map<LambdaCategory, LambdaGroup> LambdaMap;

    LambdaMap mLambdaMap;

  private:
    std::stringstream mGeneratorStream;

  private:
    LambdaMap::iterator lambdaGroup (const LambdaCategory& type);
    
    // Get the lambda group for the given number of parameters (starting
    // a new group if necessary)

    unsigned categorysize(const LambdaCategory& typeId) const;
    const std::string findFirstFunc(const LambdaCategory& typeId) const;

    void startExplicitImpl (std::stringstream &strm, const LambdaCategory& typeId);
    void writeGenerator (std::stringstream &strm, int id, const lambdaT &lam);
  };
}

#endif // CodeGen_rmg_20020211_included
